from fastapi import APIRouter, Depends, HTTPException
from starlette_context import context
import logging
from auth import auth
import conf
import opa
import correlation

router = APIRouter()

logger = logging.getLogger(__name__)
if conf.ENABLE_DEV_DIAGNOSTICS:
    from pympler.tracker import SummaryTracker

    tracker = SummaryTracker()


@router.get("/diag/require_authorized_user")
async def diag_require_authorized_user(
    auth_data: auth.Auth = Depends(auth.require_authorized_user)
):
    """
    This API requires you to be an authorized user.
    Diagnostic API - For development testing only
    """
    return {
        "groups": auth_data.groups,
        "data_partition_id": auth_data.data_partition_id,
    }


@router.get("/diag/require_authorized_admin")
async def diag_require_authorized_admin(
    auth_data: auth.Auth = Depends(auth.require_authorized_admin)
):
    """
    This API requires you to be in authorized admin (be in admin group).
    Diagnostic API - For development testing only
    """
    return {
        "groups": auth_data.groups,
        "data_partition_id": auth_data.data_partition_id,
    }


@router.get("/diag/require_authorized_ops")
async def diag_require_authorized_ops(
    auth_data: auth.Auth = Depends(auth.require_authorized_ops)
):
    """
    This API requires you to be in authorized ops (be in ops group).
    Diagnostic API - For development testing only
    """
    return {
        "groups": auth_data.groups,
        "data_partition_id": auth_data.data_partition_id,
    }


@router.get("/diag/require_data_partition_id")
async def diag_require_data_partition_id_in_header(
    data_partition_id: str = Depends(auth.require_data_partition_id)
):
    """
    This testing API requires a data partition id in the header.
    Diagnostic API - For development testing only
    """
    return {"data_partition_id": data_partition_id}


@router.get("/diag/tracker")
def diag_tracker(auth_data: auth.Auth = Depends(auth.require_authorized_admin)):
    """
    This API is used to capture changes in objects in memory and should only be used with direct backend access.
    It was added to help discover memory leaks.

    Diagnostic API - For development testing only
    """
    return {tracker.print_diff()}


@router.get("/diag/policies")
def diag_fetch_all_policies(
    auth_data: auth.Auth = Depends(auth.require_authorized_admin)
):
    """
    Return all policies from OPA directly with no filtering.
    Diagnostic API - For development testing only
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)

    opa_response = opa.fetch_opa_policies_direct()
    logger.debug(f"OPA return {opa_response.json}")
    if opa_response.ok:
        return opa_response.json
    else:
        raise HTTPException(
            status_code=opa_response.status_code, detail=opa_response.message
        )
