# policy_eval_api.py

# Copyright 2022 <XYZ>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fastapi
from fastapi import Depends, HTTPException, UploadFile, Query, Response
import sys
import os
import logging
import json
from starlette_context import context
from pydantic import BaseModel
import conf
import correlation

from starlette.status import (
    HTTP_400_BAD_REQUEST,
    HTTP_406_NOT_ACCEPTABLE,
    HTTP_422_UNPROCESSABLE_ENTITY,
    HTTP_500_INTERNAL_SERVER_ERROR,
    HTTP_503_SERVICE_UNAVAILABLE,
)

sys.path.append(os.path.abspath(".."))
import opa
import safe
from auth import auth


class Detail(BaseModel):
    detail: str


logger = logging.getLogger(__name__)
router = fastapi.APIRouter()

# Header code is compatible with Python 3.6 - 3.9, it will not work with Python 3.10
# See https://fastapi.tiangolo.com/tutorial/header-params/


@router.post(
    "/evaluations/query",
    responses={
        HTTP_400_BAD_REQUEST: {"model": Detail},
        HTTP_500_INTERNAL_SERVER_ERROR: {"model": Detail},
        HTTP_503_SERVICE_UNAVAILABLE: {"model": Detail},
    },
)
def evaluate_policy(
    response: Response,
    policy_id: str,
    file: UploadFile,
    auth_data: auth.Auth = Depends(auth.require_authorized_user),
    include_auth: bool = Query(
        default=False,
        description="Update posted data to include auth (token, xuserid and data partition id) from headers",
    ),
    cache: bool = Query(
        default=True,
        description="Use cache (if enabled) on policy service.",
    ),
):
    """
    ## Evaulate Policies
    This API is to help you evaluate policies.

    If include_auth is True, then in your file data token, xuserid and data partition id will be ignored in the file and information from the headers
    of the request will be used for this information.

    ### Example:
    For example file data for policy dataauthz.rego:
    Where XXXX is the data partition and YYYY is a legal tag
    ```json
    {
        "input": {
            "operation": "update",
            "records": [
                {
                    "id":"XXXX:test:1.4.1654807204111",
                    "kind":"XXXX:bulkupdate:test:1.1.1654807204111",
                    "legal":{
                        "legaltags":[
                            "YYYY"
                        ],
                        "otherRelevantDataCountries":["US"],
                        "status":"compliant"
                    },
                    "acls":{
                        "viewers":["data.default.viewers@XXXX.group"],
                        "owners":["data.default.owners@XXXX.group"]
                    }
                }
            ]
        }
    }
    ```

    ### Permission required to use this API:

    `policy.service.user` or `policy.service.admin`

    Setting cache to disable is intended for testing policies, in particular policies that do caching in OPA.
    """
    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    response.headers["X-Correlation-ID"] = context["correlation_id"]

    if not safe.is_safe_policy_id(policy_id):
        safe_policy_id = safe.safe_encode(policy_id)
        logger.critical(f"Error Invalid policy_id: {safe_policy_id}")
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST, detail=safe.BAD_REQUEST_DETAIL
        )

    # not expected with our current dataauthz.rego
    if policy_id.endswith(".rego"):
        policy_id = policy_id.removesuffix(".rego")

    # Try to make it easier for developers
    if "osdu/instance/" in policy_id:
        logger.debug(f"Instance policy eval {policy_id}")
    elif "osdu/partition/" not in policy_id:
        policy_id = "osdu/partition/" + auth_data.data_partition_id + "/" + policy_id

    posted_data = process_posted_data(
        file=file, include_auth=include_auth, auth_data=auth_data
    )
    result = opa.data(query=json.dumps(posted_data), path=policy_id, cache=cache)

    if result.ok:
        logging.debug(result.message)
    else:
        logging.error(result.message)
        raise HTTPException(
            status_code=result.status_code,
            detail=f"Error when talking to OPA: {result.message}",
        )

    if "result" not in result.json:
        if conf.ENABLE_DEV_DIAGNOSTICS_DUMPS:
            # save lots of debugging info
            file_path = "eval_dump_bad.json"
            with open(file_path, "w") as outfile:
                outfile.write(json.dumps(posted_data, indent=4))
            print(posted_data)

        raise HTTPException(
            status_code=HTTP_406_NOT_ACCEPTABLE,
            detail=f"Evaluate failed for policy '{policy_id}'. result: {result.json}",
            headers={"X-Error-data": json.dumps(posted_data)},
        )
    else:
        if conf.ENABLE_DEV_DIAGNOSTICS_DUMPS:
            # save lots of debugging info
            file_path = "eval_dump_input_ok.json"
            with open(file_path, "w") as outfile:
                outfile.write(json.dumps(posted_data, indent=4))

            print(posted_data)

            file_path = "eval_dump_output_ok.json"
            with open(file_path, "w") as outfile:
                outfile.write(json.dumps(result.json, indent=4))

        return result.json


def process_posted_data(file: UploadFile, include_auth: bool, auth_data: auth.Auth):
    contents = file.file.read()
    try:
        posted_data = json.loads(contents.decode("utf-8"))

        if "input" not in posted_data:
            raise HTTPException(
                status_code=HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Input not found in file data",
            )
    except json.decoder.JSONDecodeError as err:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY, detail=f"JSON error {err}"
        )
    except UnicodeDecodeError as err:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY, detail=f"JSON error {err}"
        )
    except TypeError as err:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY, detail=f"Input error {err}"
        )

    if include_auth:
        # add/update token, xuserid and data partition id to input
        posted_data["input"]["token"] = auth_data.access_token
        posted_data["input"]["xuserid"] = auth_data.x_user_id
        posted_data["input"]["datapartitionid"] = auth_data.data_partition_id

    # make sure we have them
    if "token" not in posted_data["input"]:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail="token not found in file data",
        )

    if "datapartitionid" not in posted_data["input"]:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail="datapartitionid not found in file data",
        )

    return posted_data
