import logging
import tarfile
import io
import time
import hashlib

from starlette.status import (
    HTTP_200_OK,
    HTTP_202_ACCEPTED,
    HTTP_404_NOT_FOUND,
    HTTP_503_SERVICE_UNAVAILABLE,
)

from bundles import storage

from opa_response import OpaResponse

logger = logging.getLogger(__name__)


def delete_policy(data_partition, policy_id):
    """
    Delete policy from CSP providers
    """

    response = OpaResponse()
    response.message = f"Unable to remove policy {policy_id} from bundle server"

    # Double-check to make sure we have a .rego
    if not policy_id.endswith(".rego"):
        return response

    found_policy_in_bundle = False
    bundle = bundle_filename(data_partition)
    with io.BytesIO() as outbuffer:
        with io.BytesIO() as inbuffer:
            try:
                inbuffer, _ = storage.get_storage().download_file(bundle, inbuffer)
            except Exception as err:
                logger.error(f"Error Downloading {bundle} from CSP Storage: {err}")
                response.status_code = HTTP_503_SERVICE_UNAVAILABLE
                return response
            inbuffer.seek(0)
            with tarfile.open(mode="w:gz", fileobj=outbuffer) as tarout:
                with tarfile.open(mode="r:gz", fileobj=inbuffer) as tarin:
                    for member in tarin.getmembers():
                        # Skip policy to be deleted
                        if member.name == policy_id:
                            found_policy_in_bundle = True
                        else:
                            tarout.addfile(member, tarin.extractfile(member.name))
                tarout.close()
                outbuffer.seek(0)
                if found_policy_in_bundle:
                    if storage.get_storage().upload_file(bundle, outbuffer):
                        response.message = (
                            f"Policy {policy_id} removed from bundle server"
                        )
                        # Could be 200, but 202 is proper since there could be OPA servers that may still have the content
                        response.status_code = HTTP_202_ACCEPTED  # could be 200
                        response.ok = True
                        return response
                else:
                    # Return NOT FOUND, no need to upload bundle
                    # Text should contain "not_found" as OPA includes that when talking directly
                    response.message = f"Policy {policy_id} not_found in bundle server"
                    response.status_code = HTTP_404_NOT_FOUND
    return response


def get_bundle(data_partition):
    """
    Get bundle (for backup)
    """
    bundle = bundle_filename(data_partition)
    logger.debug(f"attempting to get bundle: {bundle}")

    try:
        inbuffer = io.BytesIO()
        inbuffer, _ = storage.get_storage().download_file(bundle, inbuffer)
        inbuffer.seek(0)
        return inbuffer
    except Exception as err:
        logger.critical(f"Error Downloading {bundle} from CSP Storage: {err}")
        message = f"Unable to get bundle from bundle server: download bundle {bundle} error: {err}"
        logger.critical(message)
    return None


def put_default_bundle(data_partition, policy_list):
    """
    Get bundle (for backup)
    """
    bundle = bundle_filename(data_partition)
    logger.debug(f"attempting to get bundle: {bundle}")

    try:
        with io.BytesIO() as outbuffer:
            with tarfile.open(mode="w:gz", fileobj=outbuffer) as tarout:
                for key in policy_list:
                    info = tarfile.TarInfo(key)
                    info.size = len(policy_list[key])
                    info.mtime = int(time.time())
                    tarout.addfile(info, io.BytesIO(policy_list[key]))
                # tarout.addfile(info, io.BytesIO(policy_list[key]))
                tarout.close()
                outbuffer.seek(0)
                if storage.get_storage().upload_file(bundle, outbuffer):
                    logger.info(f"Bundle {bundle} added to bundle server")
                    return HTTP_202_ACCEPTED
                else:
                    logger.critical(f"Error uploading {bundle} to CSP Storage")
    except Exception as err:
        logger.critical(f"Error uploading {bundle} from CSP Storage: {err}")
        message = f"Unable to upload bundle {bundle} to bundle server error: {err}"
        logger.critical(message)
    return None


def put_policy(data_partition, policy_id, policy):
    """
    Put policy as bundle to CSP providers
    """
    bundle = bundle_filename(data_partition)
    logger.debug(f"bundle: {bundle}")

    response = OpaResponse()
    response.message = f"Unable to add policy {policy_id} to bundle server"
    updated_existing = False

    with io.BytesIO() as outbuffer:
        with io.BytesIO() as inbuffer:
            try:
                inbuffer, _ = storage.get_storage().download_file(bundle, inbuffer)
            except Exception as err:
                logger.critical(f"Error Downloading {bundle} from CSP Storage: {err}")
                response.status_code = HTTP_503_SERVICE_UNAVAILABLE
                response.message = f"Unable to add policy {policy_id} to bundle server: download bundle {bundle} error: {err}"
                template = "An exception of type {0} occurred. Arguments:\n{1!r}"
                message = template.format(type(err).__name__, err.args)
                logger.critical(message)
                return response
            inbuffer.seek(0)
            with tarfile.open(mode="w:gz", fileobj=outbuffer) as tarout:
                with tarfile.open(mode="r:gz", fileobj=inbuffer) as tarin:
                    for member in tarin.getmembers():
                        # Do not add previous version of existing policy (in case it existed)
                        if member.name != policy_id:
                            tarout.addfile(member, tarin.extractfile(member.name))
                        elif member.name == policy_id:
                            f = tarin.extractfile(member.name)
                            contents = f.read()
                            existing_sha256 = hashlib.sha256(contents).hexdigest()
                            updated_existing = True
                            logger.info(
                                f"Request to update existing {policy_id} {existing_sha256} in bundle"
                            )

                if updated_existing:
                    updated_sha256 = hashlib.sha256(policy).hexdigest()
                    if existing_sha256 == updated_sha256:
                        tarout.close()
                        response.ok = True
                        response.message = (
                            f"Policy {policy_id} already in bundle server"
                        )
                        response.status_code = HTTP_200_OK
                        logger.info(
                            f"Ignoring update to {policy_id}. No change detected. {existing_sha256} with {updated_sha256}"
                        )
                        return response
                    else:
                        logger.info(
                            f"Replacing {policy_id} {existing_sha256} with {updated_sha256}"
                        )
                info = tarfile.TarInfo(policy_id)
                info.size = len(policy)
                tarout.addfile(info, io.BytesIO(policy))
                tarout.close()
                outbuffer.seek(0)
                if storage.get_storage().upload_file(bundle, outbuffer):
                    response.ok = True
                    if updated_existing:
                        response.message = (
                            f"Policy {policy_id} updated in bundle server"
                        )
                        response.status_code = HTTP_202_ACCEPTED
                    else:
                        response.message = f"Policy {policy_id} added to bundle server"
                        # could be 200, but 202 is proper since there could be OPA servers that may not have the content yet
                        response.status_code = HTTP_202_ACCEPTED
                    return response
                else:
                    response.status_code = HTTP_503_SERVICE_UNAVAILABLE
                    response.message = f"Unable to add policy {policy_id} to bundle server: upload error {bundle}"
                    logger.critical(f"Error uploading {bundle} to CSP Storage")

    return response


def policy_filename(policy_id):
    return "{0}.rego".format(policy_id)


def bundle_filename(data_partition):
    if data_partition:
        logger.debug(f"partition {data_partition} bundle")
        return f"bundle-{data_partition}.tar.gz"
    else:
        logger.debug("Instance/default bundle")
        return "bundle.tar.gz"


def correct_package_declaration(data_partition, policy_id):
    return 'package osdu.partition["{0}"].{1}'.format(data_partition, policy_id[:-5])


def package_name_ok(data_partition, policy_id, policy):
    logger.debug(f"data_partition: {data_partition} policy_id: {policy_id}")
    return correct_package_declaration(data_partition, policy_id) in policy.decode(
        "utf-8"
    )
