#!/usr/bin/env python3
# OSDU Policy Service
from fastapi import FastAPI, Request, Depends

# from fastapi.responses import RedirectResponse, HTMLResponse
# from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm, HTTPBearer
# from fastapi.openapi.docs import get_swagger_ui_html
# from fastapi.openapi.utils import get_openapi
from fastapi.staticfiles import StaticFiles

# from starlette_context import context, request_cycle_context
import uvicorn
import os
import logging
import conf
from api import (
    policy_read_api,
    policy_update_api,
    policy_eval_api,
    health_api,
    compile_api,
    info_api,
    translate_api,
    diag_api,
    login_api,
    backup_api,
    bootstrap_api,
    config_api,
    tenant_api,
    validate_api,
)

# from models.translate import TranslateItem
from views import home
import _version
from auth import auth

# from fastapi.middleware.cors import CORSMiddleware
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
import correlation
import time
# import pdb; pdb.set_trace()

# TranslateItem.update_forward_refs()

tags_metadata = [
    {"name": "policies", "description": "Policy Maintenance"},
    {"name": "evaluate", "description": "Evaluate APIs"},
    {
        "name": "translate",
        "description": "Translation APIs",
        "externalDocs": {
            "description": "Translate API documentation",
            "url": "https://osdu.pages.opengroup.org/platform/security-and-compliance/policy/translate/",
        },
    },
    {
        "name": "diagnostic",
        "description": "Version information and other diagnostic APIs",
    },
    {"name": "utility", "description": "Service APIs"},
    {"name": "login", "description": "Login for Admin UI"},
    {"name": "home", "description": "Home Page"},
]

description = f"""
    OSDU Policy Service API v{_version.__apiversion__} Build {_version.__version__} for {_version.__milestone__}

Policy service is used for management and evaluation of dynamic policies in OSDU.

### Dependencies:
Policy service has a runtime dependency on the following:
* Open Policy Agent [OPA](https://openpolicyagent.org/)
* OSDU entitlements service for API authorization

### API authorization
* To read or evaluate policies, user calling the policy service, must be a member of _service.policy.user_ or _service.policy.admin_ group.
* To create, update, or delete the policies, user must be a member of _service.policy.admin_ group.
* User can be added to these groups by using entitlements service.

### Useful information
* [Documentation](https://osdu.pages.opengroup.org/platform/security-and-compliance/policy/)
* [Source](https://community.opengroup.org/osdu/platform/security-and-compliance/policy)
"""

# setup loggers and make sure we can find logging.config
log_file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "logging.conf")
logging.config.fileConfig(log_file_path, disable_existing_loggers=False)

# create formatter
logging.basicConfig(
    format="%(asctime)s logger=%(name)s loglevel=%(levelname)-6s correlation-id=%(correlation_id)s %(funcName)s() L%(lineno)-4d %(message)s   call_trace=%(pathname)s L%(lineno)-4d"
)

logging.setLogRecordFactory(correlation.set_correlation_id())

# get root logger
logger = logging.getLogger(
    __name__
)  # the __name__ resolve to "main" since we are at the root of the project.
# This will get the root logger since no logger in the configuration has this name.
logger.setLevel(conf.LOG_LEVEL)

# create console handler and set level
# ch = logging.StreamHandler()
# ch.setLevel(conf.LOG_LEVEL)

# formatter = logging.Formatter('%(asctime)s loglevel=%(levelname)-6s logger=%(name)s %(funcName)s() L%(lineno)-4d %(message)s   call_trace=%(pathname)s L%(lineno)-4d')

# add formatter to ch
# ch.setFormatter(formatter)

# add ch to logger
# logger.addHandler(ch)

logger.info(f"Log level is {logger.getEffectiveLevel()}")

# disable docs
# app = FastAPI(docs_url=None, redoc_url=None, openapi_url = None)

middleware = [
    Middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )
]

"""
def policy_context_dependency(correlation_id: str = Header(None), user_agent: str = Header(None)):
    # When used a Depends(), this fucntion get the `Correlation-ID` header,
    # which will be documented as a required header by FastAPI.
    # use `correlation_id: str = Header(None)` for an optional header.
    # (correlation_id = Header(...) for required header

    data = {"correlation_id": correlation_id,
            "user_agent": user_agent
           }
    with request_cycle_context(data):
        # yield allows it to pass along to the rest of the request
        yield
"""


app = FastAPI(
    middleware=middleware,
    title="OSDU Policy Service",
    description=description,
    version=f"{_version.__apiversion__}",
    docs_url=conf.SERVICE_BASE_PATH + "/docs",
    redoc_url=conf.SERVICE_BASE_PATH + "/redoc",
    openapi_url=conf.SERVICE_BASE_PATH + "/openapi.json",
    contact={
        "name": "OSDU Policy Service",
        "url": "https://osdu.pages.opengroup.org/platform/security-and-compliance/policy/",
    },
    license_info={
        "name": "Apache 2.0",
        "url": "https://community.opengroup.org/osdu/platform/security-and-compliance/policy/-/raw/trusted-FastAPI/LICENSE",
    },
    dependencies=[Depends(correlation.policy_context_dependency)],
    openapi_tags=tags_metadata,
)

# @app.on_event("startup")
# async def startup_event():
#    logger = logging.getLogger("uvicorn.access")
#    handler = logging.StreamHandler()
#    handler.setFormatter(logging.Formatter("%(asctime)s - %(levelname)s - %(message)s"))
#    logger.addHandler(handler)


@app.middleware("http")
async def add_process_time_header(request: Request, call_next):
    start_time = time.time()
    response = await call_next(request)
    process_time = time.time() - start_time
    response.headers["X-Process-Time"] = str(process_time)

    # ENVOY is enabled, propagate some request headers to response headers
    if conf.ENVOY:
        # x-request-id
        x_client_trace_id = request.headers.get("x-client-trace-id")

        # x-request-id
        x_request_id = request.headers.get("x-request-id")

        # B3 Propagation is a specification for the header "b3" and those that start with "x-b3-".
        # These headers are used for trace context propagation across service boundaries.
        # Most commonly used by Zipkin
        x_b3_traceid = request.headers.get("x-b3-traceid")
        x_b3_spanid = request.headers.get("x-b3-spanid")

        # OTEL / Lightstep / ServiceNow tracer
        x_ot_span_context = request.headers.get("x-ot-span-context")

        # Datadog tracer
        x_datadog_trace_id = request.headers.get("x-datadog-trace-id")

        # SkyWalking tracer
        sw8 = request.headers.get("sw8")

        # AWS X-Ray tracer
        x_amzn_trace_id = request.headers.get("x-amzn-trace-id")

        # Provide trace and request information in response headers
        if x_request_id:
            response.headers["x-request-id"] = x_request_id
        if x_client_trace_id:
            response.headers["x-client-trace-id"] = x_client_trace_id
        if x_b3_traceid:
            response.headers["x-b3-traceid"] = x_b3_traceid
        if x_b3_spanid:
            response.headers["x-b3-spanid"] = x_b3_spanid
        if x_ot_span_context:
            response.headers["x-ot-span-context"] = x_ot_span_context
        if x_datadog_trace_id:
            response.headers["x-datadog-trace-id"] = x_datadog_trace_id
        if sw8:
            response.headers["sw8"] = sw8
        if x_amzn_trace_id:
            response.headers["x-amzn-trace-id"] = x_amzn_trace_id

    return response


def configure():
    configure_routing()


def configure_routing():
    ADMIN_PROTECTED = [Depends(auth.require_authorized_admin)]
    PROTECTED = [Depends(auth.require_authorized_user)]

    app.include_router(home.router, tags=["home"])
    app.include_router(
        policy_read_api.router,
        prefix=conf.SERVICE_BASE_PATH,
        tags=["policies"],
        dependencies=PROTECTED,
    )
    app.include_router(
        policy_update_api.router,
        prefix=conf.SERVICE_BASE_PATH,
        tags=["policies"],
        dependencies=ADMIN_PROTECTED,
    )
    app.include_router(
        policy_eval_api.router,
        prefix=conf.SERVICE_BASE_PATH,
        tags=["evaluate"],
        dependencies=PROTECTED,
    )
    app.include_router(
        translate_api.router,
        prefix=conf.SERVICE_BASE_PATH,
        tags=["translate"],
        dependencies=PROTECTED,
    )
    app.include_router(
        info_api.router, prefix=conf.SERVICE_BASE_PATH, tags=["diagnostic"]
    )
    app.include_router(
        compile_api.router,
        prefix=conf.SERVICE_BASE_PATH,
        tags=["utility"],
        dependencies=PROTECTED,
    )
    app.include_router(
        tenant_api.router,
        prefix=conf.SERVICE_BASE_PATH,
        tags=["utility"],
        dependencies=ADMIN_PROTECTED,
    )
    app.include_router(
        health_api.router, prefix=conf.SERVICE_BASE_PATH, tags=["diagnostic"]
    )
    app.include_router(
        validate_api.router,
        prefix=conf.SERVICE_BASE_PATH,
        tags=["diagnostic"],
        dependencies=PROTECTED,
    )
    app.include_router(
        backup_api.router,
        prefix=conf.SERVICE_BASE_PATH,
        tags=["utility"],
        dependencies=ADMIN_PROTECTED,
    )
    app.include_router(
        bootstrap_api.router,
        prefix=conf.SERVICE_BASE_PATH,
        tags=["utility"],
        dependencies=ADMIN_PROTECTED,
    )

    if conf.ALLOW_CONFIG_API:
        app.include_router(
            config_api.router,
            prefix=conf.SERVICE_BASE_PATH,
            tags=["diagnostic"],
            dependencies=ADMIN_PROTECTED,
        )

    if conf.ENABLE_DEV_DIAGNOSTICS:
        # do not enable diags in production
        app.include_router(
            diag_api.router, tags=["diagnostic"], dependencies=ADMIN_PROTECTED
        )
        logger.warning("ENABLE_DEV_DIAGNOSTICS enabled")


def configure_middleware():
    # This will need some work before Admin UI is moved out of POC
    # origins = [
    #     "http://localhost",
    #     "http://localhost:8080",
    #     "http://0.0.0.0:8080",
    #     "http://localhost:4200"
    # ]
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )


if __name__ == "__main__":
    configure()
    # log_config = uvicorn.config.LOGGING_CONFIG
    # log_config["formatters"]["access"]["fmt"] = "%(asctime)s - %(levelname)s - %(message)s"
    # log_config["formatters"]["default"]["fmt"] = "%(asctime)s - %(levelname)s - %(message)s"
    uvicorn.run(app, port=8080, host="0.0.0.0", log_config=None)
else:
    configure()
